#!/usr/bin/env python
# filldb.py
#
#    ========== licence begin LGPL
#    Copyright (C) 2002-2006 SAP AG
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#    ========== licence end
#

import sys
import string

def fatalError (msg):
    sys.stderr.write ('filldb: ' + msg + '\n')
    sys.exit (2)

def importSQLModule ():
    try:
        import sdb.sql
        return sdb.sql
    except ImportError:
        pass
    try:
        import sapdb.sql
        return sapdb.sql
    except ImportError:
        pass
    try:
        import sapdb
        return sapdb
    except ImportError:
        pass
    fatalError ('Cannot import one of sdb.sql, sapdb.sql, sapdb')

sql = importSQLModule ()

def checkOptions (options):
    if options.data and options.log:
        fatalError ('only one of -data and -log can be specified')
    if not options.data and not options.log and not options.clear:
        fatalError ('one of -data, -log, -clear must be specified')
    if not options.userInfo:
        fatalError ('-u <username>,<password> must be specified')
    if not options.dbname:
        fatalError ('-d <dbname> must be specified')

def connect (options, userInfoName = 'userInfo',
        dbnameName = 'dbname', nodeName = 'node',
        propName = 'connectProperties', module = None):
    userInfo = getattr (options, userInfoName)
    dbname = getattr (options, dbnameName)
    node = getattr (options, nodeName)
    try:
        properties = getattr (options, propName)
    except AttributeError:
        properties = ''
    pos = string.find (userInfo, ',')
    if pos != -1:
        username = userInfo [:pos]
        pwd = userInfo [pos + 1:]
    else:
        username = userInfo
        pwd = ''
    if not node:
        pos = string.find (dbname, ':')
        if pos != -1:
            node = dbname [:pos]
            dbname = dbname [pos + 1:]
        else:
            node = ''
    if module == None:
        module = importSQLModule ()
    try:
        result = module.connect (username, pwd, dbname, node) # , properties)
    except module.SQLError, err:
        if err.errorCode == -4008:
            try:
                username = string.upper (username)
                pwd = string.upper (pwd)
                result = module.connect (username, pwd, dbname, node)
            except module.SQLError, err:
                fatalError (str (err))
        else:
            fatalError (str (err))
    return result

def dbversion (session):
    (versionString,) = session.sql ('select KERNEL from versions').next ()
    fields = string.split (versionString)
    release = tuple (map (int, string.split (fields [1], '.')))
    return release

_tablename = 'FILLTABLE'
_colSize = 1000

def dropDataTable (session):
    try:
        session.sql ('DROP TABLE ' + _tablename)
        session.commit ()
    except sql.SQLError, err:
        if err.errorCode != -4004:
            fatalError ('Dropping table: ' + str (err))

class DBFiller:
    pagename = 'pages'

    def __init__ (self, session, rowCount = 1000, commitCount = 100):
        self.session = session
        self.rowCount = rowCount
        self.commitCount = commitCount
        self.prepareStat ()

    def prepareStat (self):
        self.statSelect = self.session.prepare (self.statSelectCmd)

    def currentStat (self):
        raise 'not implemented'

    def createDataTable (self):
        try:
            self.session.sql ('exists table ' + _tablename)
            return
        except sql.SQLError, err:
            pass
        try:
            self.session.sql ('create table ' + _tablename + '('
                + 'keycol serial key, datacol varchar (%d))' % _colSize)
        except sql.SQLError, err:
            fatalError ('Creating table: ' + str (err))
        for i in xrange (1):
            try:
                self.session.sql ("""create index index%d
                    on %s (datacol, keycol)""" % (i, _tablename))
            except sql.SQLError, err:
                fatalError ('Creating index: ' + str (err))

    def commitHook (self):
        pass

    def go (self):
        self.createDataTable ()
        data = self.data
        insert = self.session.prepare ('insert into ' + _tablename + ' (datacol) values (?)')
        maxRow = self.rowCount
        rowIndex = 0
        while 1:
            insert.execute ([data])
            rowIndex = rowIndex + 1
            if rowIndex % self.commitCount == 0:
                self.session.commit ()
                sys.stdout.write ('\r%9d rows inserted, %3d %% free %s'
                    % (rowIndex, self.currentStat (), self.pageName))
                sys.stdout.flush ()
                self.commitHook ()
            if (maxRow > 0) and (rowIndex >= maxRow):
                break

class DataFiller75 (DBFiller):
    pageName = 'data pages'
    statSelectCmd = 'select PCT_PAGES_UNUSED from INFO_SYSDD_SERVERDBSTATE'
    data = 'x'

    def currentStat (self):
        cursor = self.statSelect.execute ([])
        row = cursor.next ()
        result = row [0]
        del cursor
        return result

class DataFiller (DBFiller):
    pageName = 'data pages'
    statSelectCmd = 'select 100 - (SUM (USEDSIZE) * 100 / SUM (USABLESIZE)) from DATAVOLUMES'
    data = 'x'

    def currentStat (self): 
        cursor = self.statSelect.execute ([]) 
        percentage, = cursor.next () 
        del cursor 
        return percentage 

class LogFiller75 (DBFiller):
    pageName = 'log pages'
    statSelectCmd = 'select LOG_PAGES, USED_LOG_PAGES from INFO_SYSDD_SERVERDBSTATE'
    data = 'x' * _colSize

    def currentStat (self):
        cursor = self.statSelect.execute ([])
        pages, usedPages = cursor.next ()
        del cursor
        return 100 - usedPages * 100 / pages

    def commitHook (self):
        self.session.sql ('delete from ' + _tablename + ' where keycol > 0')

class LogFiller (DBFiller):
    pageName = 'log pages'
    statSelectCmd = 'select 100 - USEDSIZEPERCENTAGE from LOGSTATISTICS'
    data = 'x' * _colSize

    def currentStat (self):
        cursor = self.statSelect.execute ([])
        percentage, = cursor.next ()
        del cursor
        return percentage

    def commitHook (self):
        self.session.sql ('delete from ' + _tablename + ' where keycol > 0')

def main (options, args):
    """
    This script inserts rows into the database, thus
    filling either the data or the log volumes.

    This is useful to demonstrate
    - how filling the data or log volumes looks like in the DBM GUI
    - the actions required when clients block because the volumes 
      are filled to the max

    Options:
    """
    checkOptions (options)
    session = connect (options, module = sql)
    version = dbversion (session)
    if version <= (7, 5, 9):
        dataFiller, logFiller = DataFiller75, LogFiller75
    else:
        dataFiller, logFiller = DataFiller, LogFiller
    try:
        if options.clear:
            dropDataTable (session)
        if options.data:
            filler = dataFiller (session, options.rowCount, options.commitCount)
            filler.go ()
        elif options.log:
            filler = logFiller (session, options.rowCount, options.commitCount)
            filler.go ()
    finally:
        del session

def _options ():
    userDefault = None # 'SAPR3,SAP'
    dbDefault = None   # 'TST'
    rowCountDefault = 1000
    return [
        # (optstring, varname, typechar, default, help)
        ('data', None, '', None, 'fill data volumes'),
        ('log', None, '', None, 'fill log volumes'),
        ('u', 'userInfo', ':', userDefault, 'user <name>,<pwd>'),
        ('d', 'dbname', ':', dbDefault, 'db name'),
        ('n', 'node', ':', '', 'host name'),
        # ('P', 'connectProperties', '::', None, 'connect properties'),
        ('R', 'rowCount', ':', rowCountDefault, 'number of rows to insert\n\t\t0: no limit, abort with Ctrl-C'),
        ('C', 'commitCount', ':', 100, 'number of commands between COMMIT'),
        ('clear', None, '', None, 'drop the table'),
        ]

if __name__ == '__main__':
    import optlib
    optlib.optMain2 (main, _options ())
